module bi_ari; 


{============================================================================}
{===  mpi_ari  ======   Bounded Interval  Arithmetic  =======================}
{====================   =       =         ===         =======================}
{============================================================================}
{                                                                            }
{    File    :  bi_ari.p                                                     }
{                                                                            }
{=   Version :  1.1                                                          }
{                                                                            }
{=   Date    :  05/../..  ( first implementation )                           }
{=              ../../..  ( last change )                                    }
{                                                                            }
{    Language:  PASCAL-XSC, Vers. 3.0 or later                               }
{                                                                            }
{    Author  :  K. Petras                                                    }
{                                                                            }
{============================================================================}
{      Das Modul behandelt komplexe Rechtecke um Intervalle                  }
{      Es  rechnet gleichzeitig mit Intervallen und solchen                  }
{      zur reellen Achse symmetrischen Rechtecken und prueft                 }
{      jeweils, ob die Funktion dort analytisch ist.                          }
{============================================================================}

 use global i_ari; 


global type b_interval = global record
 int : interval;
   b : boolean; end;

  
global function construct(x : interval) : b_interval;
begin
   construct.int := x;
   construct.b   := true;
end; { construct }


global operator := (var n : integer;
		    b : boolean ); 
begin
   if (b) then
      n:=1
   else
      n:=0;
end;

global function max(x,y :real ): real; 
begin
   if (x>y) then
      max:=x
   else
      max:=y;
end;

global function min(x,y :integer):integer; 
begin
   if (x>y) then
      min:=y
   else
      min:=x;
end;



{==========   Vorzeichenumkehr =================================}

global operator -  ( x : b_interval) ch : b_interval;
begin
   ch.int:=-x.int;
   ch.b:=x.b;
end;

{==========   Addition  =======================================}

global operator +  ( x,y : b_interval) add : b_interval;
begin
   if (x.b and y.b) then
   begin
      add.int := x.int + y.int; 
      add.b    := true;
   end
   else
   begin
      add.int := 0; 
      add.b   := false;
   end;
end; { add }



global operator +  (x : b_interval;
		    y : interval ) add : b_interval;
begin
   add := x+construct(y); 
end; { add }

global operator +  (x : interval;
		    y : b_interval ) add : b_interval;
begin
   add := y+construct(x); 
end; { add }



global operator +  (x : real;
		    y : b_interval ) add : b_interval;
begin
   add:=intval(x)+y;
end; { ade_neu }

global operator +  (x : integer;
		    y : b_interval ) add : b_interval;
begin
   add:=intval(x)+y;
end;

global operator +  ( y : b_interval; x : real) add : b_interval;
begin
   add:=intval(x)+y;
end; { ade_neu }

global operator +  ( y : b_interval; x : integer) add : b_interval;
begin
   add:=intval(x)+y;
end; { ade_neu }







{==========   Subtraktion  =======================================}

global operator -  ( x,y : b_interval) diff : b_interval;
begin
   if (x.b and y.b) then
   begin
      diff.int := x.int - y.int; 
      diff.b    := true;
   end
   else
   begin
      diff.int := 0; 
      diff.b   := false;
   end;
end; { diff }



global operator -  (x : b_interval;
		    y : interval ) diff : b_interval;
begin
   diff := x-construct(y); 
end; { diff }

global operator -  (x : interval;
		    y : b_interval ) diff : b_interval;
begin
   diff := construct(x)-y; 
end; { diff }

     
global operator -  (x : real;
		    y : b_interval ) sub : b_interval;
begin
   sub:=intval(x)-y;
end; { ade_neu }

operator -  (x : integer;  y : b_interval ) sub : b_interval;
begin
   sub:=intval(x)-y;
end;

global operator -  ( y : b_interval; x : real) sub : b_interval;
begin
   sub:=y-intval(x);
end; { ade_neu }

global operator -  ( y : b_interval; x : integer) sub : b_interval;
begin
   sub:=y-intval(x);
end; { ade_neu }




{==========   Multiplikation  =======================================}

global operator *  ( x,y : b_interval) prod : b_interval;
begin
   if (x.b and y.b) then
   begin
      prod.int := x.int * y.int; 
      prod.b    := true;
   end
   else
   begin
      prod.int := 0; 
      prod.b   := false;
   end;
end; { prod }



global operator *  (x : b_interval;
		    y : interval ) prod : b_interval;
begin
   prod := x*construct(y); 
end; { prod }

global operator *  (x : interval;
		    y : b_interval ) prod : b_interval;
begin
   prod := y*construct(x); 
end; { prod }
global operator *  (x : real;
		    y : b_interval ) mul : b_interval;
begin
   mul:=intval(x)*y;
end; { ade_neu }

 operator *  (x : integer;
		    y : b_interval ) mul : b_interval;
begin
   mul:=intval(x)*y;
end;

global operator *  ( y : b_interval; x : real) mul : b_interval;
begin
   mul:=y*intval(x);
end; { ade_neu }

global operator *  ( y : b_interval; x : integer) mul : b_interval;
begin
   mul:=y*intval(x);
end; { ade_neu }



{==========   Division  =======================================}

global operator /  ( x,y : b_interval) quot : b_interval;
begin
   if (0 in y.int) then
   begin
      quot.int := 0;
      quot.b   := false;
   end
   else
      if (x.b and y.b) then
      begin
	 quot.int := x.int / y.int; 
	 quot.b    := true;
      end
      else
      begin
	 quot.int := 0; 
	 quot.b   := false;
      end;
end; { quot }


global operator /  (x : b_interval;
		    y : interval ) quot : b_interval;
begin
   quot := x/construct(y); 
end; { quot }

global operator /  (x : interval;
		    y : b_interval ) quot : b_interval;
begin
   quot := y/construct(x); 
end; { quot }



global operator /  (x : real;
		    y : b_interval ) quot : b_interval;
begin
   quot:=intval(x)/y;
end; { ade_neu }

operator /  (x : integer;
		    y : b_interval ) quot : b_interval;
begin
   quot:=intval(x)/y;
end;

global operator /  ( y : b_interval; x : real) quot : b_interval;
begin
   quot:=y/intval(x);
end; { ade_neu }

global operator /  ( y : b_interval; x : integer) quot : b_interval;
begin
   quot:=y/intval(x);
end; { ade_neu }




{==========   Standardfunktionen :     ======================}

	 
global function sqr (x	: b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      if (expo(sup(x.int))>511) then
      begin
	 sqr.int := 0;
	 sqr.b   := false;
      end
      else
      begin
	 sqr.int := sqr(x.int); 
	 sqr.b   := true;
      end
   end
   else
   begin
      sqr.int := 0;
      sqr.b   := false;
   end;
end; { sqr }

global function pow (  x: b_interval;  n:integer) : b_interval;
var xq	    : interval;
   hilf	    : real;
   k, i, si : integer;
   pot2	    : array[1..1030] of interval; 
   prod	    : interval; 
begin
   if (not(x.b)) then
   begin
      pow.b := false;
      pow.int := intval(0);
   end
   else
      if (n<0) then
	 pow:=intval(1.0)/pow(x, -n)
      else
      begin
	 if (n=0) then
	 begin
	    if (0 in x.int) then         {===== Ausdruck 0 hoch 0 ======}
	    begin
	       pow.b    := false; 
	       pow.int := intval(0); 
	    end
	    else
	    begin
	       pow.int := intval(1);
	       pow.b    := true; 
	    end
	 end
	 else
	 begin                  { ===========  n>0 =====================}
	    if (n*max(expo(sup(x.int)), expo(inf(x.int))) >1020) then
	    begin
	       pow.int := intval(0) ;
	       pow.b   := false;
	    end
	    else
	    begin
	       pow.b := true; 
	       k:=2;
	       i:=1;
	       pot2[1] := x.int;
	       while (k<=n) do
	       begin
		  i:=i+1; 
		  pot2[i]:=sqr(pot2[i-1]);
		  k:=2*k;
	       end;
	    end;
	    k:=k div 2;
	    prod:=pot2[i];
	    si:=k; 
	    while (si<n) do
	    begin
	       k:=k div 2;
	       i:=i-1; 
	       if (si+k<=n) then
	       begin
		  prod:=prod*pot2[i];
		  si:=si+k;
	       end;
	    end;
	    pow.int:=prod; 
	 end;
      end;
end;
   

global function pow (  x: interval;  n:integer) : interval;
var xq	    : interval;
   hilf	    : real;
   k, i, si : integer;
   pot2	    : array[1..1030] of interval; 
   prod	    : interval; 
begin
   if (n<0) then
   begin
	 pow:=intval(1.0)/pow(x, -n)
   end
   else
   begin
      if (n=0) then
      begin
	 if (0 in x) then         {===== Ausdruck 0 hoch 0 ======}
	 begin
	    writeln(' Fehler :  Ausdruck 0 hoch 0 '); 
	    pow := intval(0); 
	 end
	 else
	 begin
	    pow := intval(1);
	 end
      end
      else
      begin                  { ===========  n>0 =====================}
	 k:=2;
	 i:=1;
	 pot2[1] := x;
	 while (k<=n) do
	 begin
	    i:=i+1; 
	    pot2[i]:=sqr(pot2[i-1]);
	    k:=2*k;
	 end;
      end;
      k:=k div 2;
      prod:=pot2[i];
      si:=k; 
      while (si<n) do
      begin
	 k:=k div 2;
	 i:=i-1; 
	 if (si+k<=n) then
	 begin
	    prod:=prod*pot2[i];
	    si:=si+k;
	 end;
      end;
      pow:=prod; 
   end;
end; { pow }
   

 	 
global function pospart (x : b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      if (0>=sup(x.int)) then
      begin
	 pospart.int := 0;
	 pospart.b   := true;
      end
      else
      begin
	 if (0 <= inf(x.int)) then
	    pospart:=x
	 else
	 begin
	    pospart.int := 0.0 +* x.int;
	    pospart.b   := true;
	 end
      end
   end
   else
   begin
      pospart.int := 0;
      pospart.b   := false;
   end;
end; { pospart }

 
	 
global function exp (x	: b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      if (sup(x.int)>709) then
      begin
	 exp.int := 0;
	 exp.b   := false;
      end
      else
      begin
	 exp.int := exp(x.int); 
	 exp.b   := true;
      end
   end
   else
   begin
      exp.int := 0;
      exp.b   := false;
   end;
end; { exp }

 

global function sin (x	: b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      sin.int := sin(x.int); 
      sin.b   := true;
   end
   else
   begin
      sin.int := 0;
      sin.b   := false;
   end
end; { sin }

global function cos (x	: b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      cos.int := cos(x.int); 
      cos.b   := true;
   end
   else
   begin
      cos.int := 0;
      cos.b   := false;
   end
end; { cos }

global function sqrtabs (x : b_interval) : b_interval;
begin 
   if (x.b) then
   begin
      sqrtabs.int := sqrt(abs(x.int)); 
      sqrtabs.b   := true;
   end
   else
   begin
      sqrtabs.int := 0;
      sqrtabs.b   := false;
   end
end; 

global function sqrtabs (x : interval) : interval;
begin 
   sqrtabs := sqrt(abs(x)); 
end; 

global function abs (x: b_interval) : b_interval;
begin
   abs.int := abs(x.int);
   abs.b := x.b;
end; { abs }






   

end.


